"""
Git operations for cloning and validating user repositories.
"""

import os
import platform
import shutil
from enum import Enum
from pathlib import Path
from typing import Optional

import git

from readmeai._exceptions import GitCloneError
from readmeai.core.logger import Logger

_logger = Logger(__name__)


class GitHost(str, Enum):
    """
    Enum class for Git service providers. Enum data includes the following:
        - domain name of the Git service
        - api url to fetch repository details
        - file url to format links in the generated README.md file
    """

    LOCAL = "local"
    GITHUB = "github.com"
    GITLAB = "gitlab.com"
    BITBUCKET = "bitbucket.org"

    @property
    def api_url(self):
        """Gets the API URL for the Git service."""
        api_urls = {
            "local": None,
            "github.com": "https://api.github.com/repos/",
            "gitlab.com": "https://api.gitlab.com/v4/projects/",
            "bitbucket.org": "https://api.bitbucket.org/2.0/repositories/",
        }
        return api_urls[self.value]

    @property
    def file_url_template(self):
        """Gets the file URL template for accessing files on the Git service."""
        file_url_templates = {
            "local": "{file_path}",
            "github.com": "https://github.com/{full_name}/blob/master/{file_path}",
            "gitlab.com": "https://gitlab.com/{full_name}/-/blob/master/{file_path}",
            "bitbucket.org": "https://bitbucket.org/{full_name}/src/master/{file_path}",
        }
        return file_url_templates[self.value]


async def clone_repository(repository: str, temp_dir: str) -> str:
    """Clone repository to temporary directory and return the path."""
    try:
        temp_dir_path = Path(temp_dir)
        if not temp_dir_path.exists():
            temp_dir_path.mkdir(parents=True)

        if Path(repository).is_dir():
            repo = git.Repo.init(temp_dir)
            origin = repo.create_remote(
                "origin", f"file://{Path(repository).absolute()}"
            )
            repo.git.config("core.sparseCheckout", "true")

            sparse_checkout_path = (
                temp_dir_path / ".git" / "info" / "sparse-checkout"
            )
            with sparse_checkout_path.open("w") as sc_file:
                sc_file.write("/*\n!.git/\n")

            origin.fetch()
            repo.git.checkout("FETCH_HEAD")
        else:
            git.Repo.clone_from(
                repository, temp_dir, depth=1, single_branch=True
            )

        await remove_hidden_contents(temp_dir_path)

        return temp_dir

    except (
        git.GitCommandError,
        git.InvalidGitRepositoryError,
        git.NoSuchPathError,
    ) as exc:
        raise GitCloneError(f"Error cloning repository: {str(exc)}") from exc


async def remove_hidden_contents(directory: Path) -> None:
    """Remove hidden files and directories from a specified directory."""
    for item in directory.iterdir():
        if item.name.startswith(".") and item.name != ".github":
            if item.is_dir():
                shutil.rmtree(item)
            else:
                item.unlink()


async def fetch_git_api_url(repo_url: str) -> str:
    """Parses the repository URL and returns the API URL."""
    try:
        parts = repo_url.rstrip("/").split("/")
        repo_name = f"{parts[-2]}/{parts[-1]}"
        for service in GitHost:
            if service in repo_url:
                api_url = f"{service.api_url}{repo_name}"
                _logger.info(f"{service.name.upper()} API URL: {api_url}")
                return api_url

        raise ValueError("Unsupported Git service.")

    except (IndexError, ValueError) as exc:
        raise ValueError(f"Invalid repository URL: {repo_url}") from exc


def fetch_git_file_url(file_path: str, full_name: str, repo_url: str) -> str:
    """Returns the URL of the file in the remote repository."""
    if Path(repo_url).exists():
        return GitHost.LOCAL.file_url_template.format(file_path=file_path)

    for service in GitHost:
        if service in repo_url:
            return service.file_url_template.format(
                full_name=full_name, file_path=file_path
            )

    return file_path


def find_git_executable() -> Optional[Path]:
    """Find the path to the git executable, if available."""
    try:
        git_exec_path = os.environ.get("GIT_PYTHON_GIT_EXECUTABLE")
        if git_exec_path:
            return Path(git_exec_path)

        # For Windows, set default location of git executable.
        if platform.system() == "Windows":
            default_windows_path = Path("C:\\Program Files\\Git\\cmd\\git.EXE")
            if default_windows_path.exists():
                return default_windows_path

        # For other OS, set executable path from PATH environment variable.
        paths = os.environ["PATH"].split(os.pathsep)
        for path in paths:
            git_path = Path(path) / "git"
            if git_path.exists():
                return git_path

        return None

    except Exception as exc:
        raise ValueError("Error finding Git executable") from exc


def validate_file_permissions(temp_dir: Path) -> None:
    """Validates file permissions of the cloned repository."""
    try:
        if platform.system() != "Windows":
            permissions = temp_dir.stat().st_mode & 0o777
            if permissions != 0o700:
                raise SystemExit(
                    f"Invalid file permissions for {temp_dir}.\n"
                    f"Expected 0o700, but found {oct(permissions)}."
                )

    except Exception as exc:
        raise ValueError(
            f"Error validating file permissions: {str(exc)}"
        ) from exc


def validate_git_executable(git_exec_path: str) -> None:
    """Validate the path to the git executable."""
    try:
        if not git_exec_path or not Path(git_exec_path).exists():
            raise ValueError(f"Git executable not found at {git_exec_path}")

    except Exception as exc:
        raise ValueError("Error validating Git executable path") from exc



"""Command-line interface options for the readme-ai application."""

from __future__ import annotations

from enum import Enum
from typing import Optional

import click


class BadgeOptions(str, Enum):
    """
    Enum for CLI options for README file badge icons.
    """

    DEFAULT = "default"
    FLAT = "flat"
    FLAT_SQUARE = "flat-square"
    FOR_THE_BADGE = "for-the-badge"
    PLASTIC = "plastic"
    SKILLS = "skills"
    SKILLS_LIGHT = "skills-light"
    SOCIAL = "social"


class ImageOptions(str, Enum):
    """
    Enum for CLI options for README file header images.
    """

    # Custom image options
    CUSTOM = "custom"
    LLM = "llm"

    # Default image options
    BLACK = "https://img.icons8.com/external-tal-revivo-regular-tal-revivo/96/external-readme-is-a-easy-to-build-a-developer-hub-that-adapts-to-the-user-logo-regular-tal-revivo.png"
    BLUE = "https://raw.githubusercontent.com/PKief/vscode-material-icon-theme/ec559a9f6bfd399b82bb44393651661b08aaf7ba/icons/folder-markdown-open.svg"
    CLOUD = "https://cdn-icons-png.flaticon.com/512/6295/6295417.png"
    GRADIENT = "https://img.icons8.com/?size=512&id=55494&format=png"
    GREY = "https://img.icons8.com/external-tal-revivo-filled-tal-revivo/96/external-markdown-a-lightweight-markup-language-with-plain-text-formatting-syntax-logo-filled-tal-revivo.png"
    PURPLE = "https://img.icons8.com/external-tal-revivo-duo-tal-revivo/100/external-markdown-a-lightweight-markup-language-with-plain-text-formatting-syntax-logo-duo-tal-revivo.png"


class ModelOptions(str, Enum):
    """
    Enum for CLI options for the LLM API key.
    """

    OFFLINE = "OFFLINE"
    OLLAMA = "OLLAMA"
    OPENAI = "OPENAI"
    VERTEX = "VERTEX"


def prompt_for_image(
    context: Optional[click.Context],
    parameter: Optional[click.Parameter],
    value: Optional[str],
) -> str:
    """Prompt the user for a custom image URL."""
    if value == ImageOptions.CUSTOM.name:
        return click.prompt("Provide an image file path or URL")
    elif value == ImageOptions.LLM.name:
        return ImageOptions.LLM.value
    elif value in ImageOptions.__members__:
        return ImageOptions[value].value
    else:
        raise click.BadParameter(f"Invalid image provided: {value}")


alignment = click.option(
    "-a",
    "--alignment",
    type=click.Choice(["center", "left"], case_sensitive=False),
    default="center",
    help="Alignment for the README.md file header sections.",
)

api = click.option(
    "--api",
    type=click.Choice(
        [opt.value for opt in ModelOptions], case_sensitive=False
    ),
    default=None,
    help="""LLM service to use for generating the README.md file. The following services are currently supported:\n
    - OFFLINE # Offline mode - no LLM service used \n
    - OLLAMA  # Ollama - llama2 \n
    - OPENAI  # OpenAI - gpt-3.5-turbo \n
    - VERTEX  # Google Cloud Vertex AI - gemini-pro) \n
    """,
)

badge_color = click.option(
    "--badge-color",
    type=str,
    default="0080ff",
    help="Custom color for the badge icon. Provide a valid color name or hex code.",
)

badge_style = click.option(
    "--badge-style",
    type=click.Choice(
        [opt.value for opt in BadgeOptions], case_sensitive=False
    ),
    default=BadgeOptions.DEFAULT.value,
    help="""\
        Badge icon style types to select from when generating README.md badges. The following options are currently available:\n
        - default \n
        - flat \n
        - flat-square \n
        - for-the-badge \n
        - plastic \n
        - skills \n
        - skills-light \n
        - social \n
        """,
)

base_url = click.option(
    "--base-url",
    type=str,
    default="https://api.openai.com/v1/chat/completions",
    help="Base URL for the LLM API service used to generate text for the README.md file.",
)

context_window = click.option(
    "--context-window",
    default=3999,
    type=int,
    help="Maximum number of tokens to use for the model's context window.",
)

emojis = click.option(
    "-e",
    "--emojis",
    is_flag=True,
    default=False,
    help="This option adds emojis to the README.md file's header sections. For example, the default header for the 'Overview' section generates the markdown code as '## Overview'. Adding the --emojis flag generates the markdown code as '## 📍 Overview'.",
)

image = click.option(
    "-i",
    "--image",
    type=click.Choice(
        [opt.name for opt in ImageOptions], case_sensitive=False
    ),
    default=ImageOptions.BLUE.name,
    callback=prompt_for_image,
    show_choices=True,
    help="""\
        Project logo image displayed in the README file header. The following options are currently supported:\n

        Custom image options:\n
        - CUSTOM (use a custom image file path or URL) \n
        - LLM (use LLM multi-modal capabilities to generate an image) \n

        Default image options:\n
        - BLACK \n
        - BLUE \n
        - CLOUD \n
        - GRADIENT \n
        - GREY \n
        - PURPLE \n
        """,
)

language = click.option(
    "-l",
    "--language",
    default="en",
    help="Language to use for generating the README.md file. Default is English (en).",
)

model = click.option(
    "-m",
    "--model",
    default="gpt-3.5-turbo",
    help="Large language model (LLM) API used to generate text for the README.md file. Default model uses OpenAI's gpt-3.5-turbo.",
)

output = click.option(
    "-o",
    "--output",
    default="readme-ai.md",
    help="Output file name for your README file. Default name is 'readme-ai.md'.",
)

rate_limit = click.option(
    "--rate-limit",
    default=5,
    type=click.IntRange(1, 20, clamp=True),
    help="Rate limit for the number of API requests per minute.",
)

repository = click.option(
    "-r",
    "--repository",
    required=True,
    help="Provide a remote repository URL (GitHub, GitLab, BitBucket), or a local directory path to your project.",
)

temperature = click.option(
    "-t",
    "--temperature",
    default=0.9,
    type=click.FloatRange(0.0, 2.0, clamp=True),
    help="Setting the model's temperature to a higher value will yield more creative content generated, while a lower value will generate more predictable content.",
)

template = click.option(
    "--template",
    type=str,
    help="README template file to use for generating the README.md file.",
)

top_p = click.option(
    "--top-p",
    default=0.9,
    type=click.FloatRange(0.0, 1.0, clamp=True),
    help="Top-p sampling probability for the model's generation. This value can be set between 0.0 and 1.0.",
)

tree_depth = click.option(
    "--tree-depth",
    default=2,
    type=int,
    help="Maximum depth of the directory tree thats included in the README.md file.",
)



"""Generates a directory tree structure for a code repository."""

from pathlib import Path


class TreeGenerator:
    """Generates a directory tree structure for a code repository."""

    def __init__(
        self, repo_name: str, root_dir: Path, repo_url: Path, max_depth: int
    ):
        self.repo_name = repo_name
        self.root_dir = root_dir
        self.repo_url = repo_url
        self.max_depth = max_depth

    def _build_tree(
        self,
        directory: Path,
        prefix: str = "",
        is_last: bool = True,
        depth: int = 0,
    ) -> str:
        """Generates a tree structure for a given directory."""
        if depth > self.max_depth:
            return ""

        children = sorted(directory.iterdir()) if directory.is_dir() else []
        children = [child for child in children]

        if not children and directory.is_dir():
            return ""

        """
        dir_name = directory.name
        if depth == 2 and directory.is_dir():
            dir_name += "               ► summary of directory"
        parts = [f"{prefix}{'└── ' if is_last else '├── '}{dir_name}"]
        """
        parts = [f"{prefix}{'└── ' if is_last else '├── '}{directory.name}"]

        for index, child in enumerate(children):
            child_prefix = prefix + ("    " if is_last else "│   ")
            child_tree = self._build_tree(
                child, child_prefix, index == len(children) - 1, depth + 1
            )

            if child_tree:
                parts.append(child_tree)

        return "\n".join(parts)

    def tree(self) -> str:
        """Generates and formats a tree structure."""
        md_tree = self._build_tree(self.root_dir)
        formatted_md_tree = md_tree.replace(
            self.root_dir.name, f"{self.repo_name}/"
        )
        return formatted_md_tree



"""Data models and functions for configuring the readme-ai CLI tool."""

from __future__ import annotations

from pathlib import Path
from typing import Optional, Union

from pydantic import BaseModel, DirectoryPath, HttpUrl, validator

from readmeai._exceptions import FileReadError
from readmeai.config.validators import GitValidator
from readmeai.core.logger import Logger
from readmeai.utils.file_handler import FileHandler
from readmeai.utils.resource_loader import get_resource_path

_logger = Logger(__name__)


class APISettings(BaseModel):
    """
    Universal LLM API settings.
    """

    content: Optional[str]
    rate_limit: Optional[int]


class FileSettings(BaseModel):
    """File paths used by the readme-ai CLI tool."""

    blacklist: str
    commands: str
    languages: str
    markdown: str
    parsers: str
    prompts: str
    shields_icons: str
    skill_icons: str


class GitSettings(BaseModel):
    """User repository settings, sanitized and validated by Pydantic."""

    repository: Union[HttpUrl, DirectoryPath]
    full_name: Optional[str]
    host: Optional[str]
    host_domain: Optional[str]
    name: Optional[str]

    _validate_repository = validator("repository", pre=True, always=True)(
        GitValidator.validate_repository
    )
    _validate_full_name = validator("full_name", pre=True, always=True)(
        GitValidator.validate_full_name
    )
    _set_host = validator("host", pre=True, always=True)(GitValidator.set_host)
    _set_name = validator("name", pre=True, always=True)(GitValidator.set_name)


class MarkdownSettings(BaseModel):
    """Markdown template blocks for the README.md file."""

    alignment: str
    badge_color: str
    badge_style: str
    badge_icons: str
    contribute: str
    emojis: bool
    features: str
    header: str
    image: str
    modules: str
    modules_widget: str
    overview: str
    placeholder: str
    quickstart: str
    shields_icons: str
    skill_icons: str
    slogan: str
    tables: str
    toc: str
    tree: str
    tree_depth: int


class ModelSettings(BaseModel):
    """LLM API settings used for generating text for the README.md file."""

    api: str
    base_url: Optional[HttpUrl]
    context_window: Optional[int]
    encoder: Optional[str]
    model: Optional[str]
    temperature: Optional[float]
    tokens: Optional[int]
    top_p: Optional[float]


class Settings(BaseModel):
    """Nested data model to store all configuration settings."""

    api: APISettings
    files: FileSettings
    git: GitSettings
    llm: ModelSettings
    md: MarkdownSettings

    class Config:
        """Pydantic configuration settings."""

        validate_assignment = True


class ConfigLoader:
    """Loads the configuration settings for the CLI."""

    def __init__(
        self,
        config_file: Union[str, Path] = "config.toml",
    ) -> None:
        """Initialize ConfigLoader with the base configuration file."""
        self.file_handler = FileHandler()
        self.config_file = config_file
        self.config = self._load_base_config()
        self._load_all_configs()

    def _load_base_config(self) -> Settings:
        """Load the base configuration file."""
        config_dict = get_resource_path(self.file_handler, self.config_file)
        return Settings.parse_obj(config_dict)

    def _load_all_configs(self) -> None:
        """Load additional configuration files specified in the Settings."""
        for (
            key,
            file_name,
        ) in self.config.files.dict().items():
            if not file_name.endswith(".toml"):
                continue

            try:
                config_data = get_resource_path(self.file_handler, file_name)
                setattr(self, key, config_data)
                _logger.debug(f"Loaded config file: {file_name}")

            except FileReadError as exc:
                setattr(self, key, None)
                _logger.warning(f"Config file not found: {file_name} - {exc}")



"""File for parsing json dependency files."""

import json
import re
from typing import List

from readmeai.core.parsers import BaseFileParser


class PackageJsonParser(BaseFileParser):
    """Parser for package.json dependency files."""

    def __init__(self) -> None:
        """Initializes the handler with given configuration."""
        super().__init__()

    def parse(self, content: str) -> List[str]:
        """Returns a list of dependencies parsed from a json file."""
        try:
            data = json.loads(content)
            package_names = []
            for section in [
                "dependencies",
                "devDependencies",
                "peerDependencies",
            ]:
                if section in data:
                    package_names.extend(data[section].keys())
            return package_names

        except json.JSONDecodeError as exc:
            return self.handle_parsing_error(f"package.json: {str(exc)}")


class YarnLockParser(BaseFileParser):
    """Parser for yarn.lock dependency files."""

    def parse(self, content: str) -> List[str]:
        """Extracts package names from a yarn.lock file."""
        try:
            return re.findall(r"(\S+)(?=@)", content)
        except re.error as exc:
            return self.handle_parsing_error(f"yarn.lock: {str(exc)}")



"""
Tokenizer utilities for tokenizing and truncating text.
"""

from tiktoken import get_encoding

from readmeai.config.settings import Settings
from readmeai.core.logger import Logger

_encoding_cache = {}
_logger = Logger(__name__)


def _set_encoding_cache(encoding_name: str) -> str:
    """Set the encoding cache for a specific encoding."""
    if encoding_name not in _encoding_cache:
        _encoding_cache[encoding_name] = get_encoding(encoding_name)
    return _encoding_cache[encoding_name]


async def token_handler(
    config: Settings, index: str, prompt: str, tokens: int
) -> str:
    """Handle token count for the prompt."""
    encoder = config.llm.encoder
    max_count = config.llm.context_window
    token_count = count_tokens(prompt, encoder)

    if token_count > max_count:
        _logger.debug(
            f"Truncating '{index}' prompt: {token_count} > {max_count} tokens!"
        )
        prompt = truncate_tokens(encoder, prompt, tokens)

    return prompt


def count_tokens(text: str, encoder: str) -> int:
    """Return the number of tokens in a text string."""
    encoding = _set_encoding_cache(encoder)
    try:
        token_count = len(encoding.encode(text, disallowed_special=()))

    except Exception as exc:
        _logger.error(
            f"Error counting tokens for '{text}' with {encoder}: {exc}"
        )
        token_count = 0

    return token_count


def truncate_tokens(encoder: str, text: str, max_count: int) -> str:
    """Truncate a text string to a maximum number of tokens."""
    if not text:
        return text
    try:
        encoder = _set_encoding_cache(encoder)
        token_count = len(encoder.encode(text))
        if token_count <= max_count:
            return text

        char_total = len(text)
        chars_per_token = char_total / token_count
        truncated_total = int(chars_per_token * max_count)
        return text[:truncated_total]

    except Exception as exc:
        _logger.error(f"Error truncating tokens for '{text}': {exc}")
        return text


def update_max_tokens(
    max_tokens: int, prompt: str, target: str = "Hello!"
) -> int:
    """Adjust the maximum number of tokens based on the specific prompt."""
    is_valid_prompt = prompt.strip().startswith(target.strip())
    return max_tokens if is_valid_prompt else max_tokens // 2



"""Functions for building and formatting the badges in the README.md file."""

from typing import Tuple

from readmeai.cli.options import BadgeOptions
from readmeai.config.settings import ConfigLoader
from readmeai.services.git import GitHost
from readmeai.utils.file_handler import FileHandler
from readmeai.utils.resource_loader import get_resource_path

_package = "readmeai.generators"
_submodule = "assets"


def _format_badges(badges: list[str]) -> str:
    """Format SVG badge icons as HTML."""
    total = len(badges)
    if badges is None or total == 0:
        return ""

    badges_per_line = total if total < 9 else (total // 2) + (total % 2)

    lines = []
    for i in range(0, total, badges_per_line):
        line = "\n\t".join(
            [
                f'<img src="{badge}" alt="{badge.split("/badge/")[1].split("-")[0]}">'
                for badge in badges[i : i + badges_per_line]
            ]
        )
        lines.append(
            f"{line}\n\t<br>" if i + badges_per_line < total else f"{line}\n"
        )

    return "\n\t".join(lines)


def build_default_badges(
    config: ConfigLoader, full_name: str, host: str
) -> str:
    """Build metadata badges using shields.io."""
    return config.md.shields_icons.format(
        host=host,
        full_name=full_name,
        badge_color=config.md.badge_color,
        badge_style=config.md.badge_style,
    )


def build_project_badges(
    dependencies: list[str], icons: dict[str, str], style: str
) -> str:
    """Build HTML badges for project dependencies."""
    badges = [
        icons[str(dependency).lower()]
        for dependency in dependencies
        if str(dependency).lower() in icons
    ]

    # Sort badges by hex value (from light to dark color)
    badges.sort(key=lambda b: int(b[1], 16) if b[1] else 0, reverse=True)
    badges = [badge[0].format(style) for badge in badges]
    return _format_badges(badges)


def shields_icons(
    conf: ConfigLoader, dependencies: list, full_name: str, git_host: str
) -> Tuple[str, str]:
    """
    Generates badges for the README using shields.io icons.
    """
    icons_dict = get_resource_path(
        FileHandler(),
        conf.files.shields_icons,
        _package,
        _submodule,
    )

    default_badges = build_default_badges(conf, full_name, git_host)

    project_badges = build_project_badges(
        dependencies, icons_dict, conf.md.badge_style
    )
    project_badges = conf.md.badge_icons.format(
        alignment=conf.md.alignment, badge_icons=project_badges
    )

    if (
        conf.md.badge_style == BadgeOptions.DEFAULT.value
        and git_host != GitHost.LOCAL
    ):
        return (
            default_badges,
            "<!-- default option, no dependency badges. -->\n",
        )

    if git_host == GitHost.LOCAL:
        return (
            "<!-- local repository, no metadata badges. -->\n",
            project_badges,
        )

    return default_badges, project_badges


def skill_icons(conf: ConfigLoader, dependencies: list) -> str:
    """
    Generates badges for the README using skill icons, from the
    repository - https://github.com/tandpfun/skill-icons.
    """
    dependencies.extend(["md"])

    icons_dict = get_resource_path(
        FileHandler(), conf.files.skill_icons, _package, _submodule
    )

    skill_icons = [
        icon for icon in icons_dict["icons"]["names"] if icon in dependencies
    ]
    skill_icons = ",".join(skill_icons)
    skill_icons = icons_dict["url"]["base_url"] + skill_icons

    if conf.md.badge_style == "skills-light":
        skill_icons = f"{skill_icons}&theme=light"

    conf.md.skill_icons = conf.md.skill_icons.format(skill_icons)

    return conf.md.badge_icons.format(
        alignment=conf.md.alignment, badge_icons=conf.md.skill_icons
    )



"""Builds each section of the README Markdown file."""

__package__ = "readmeai"

from pathlib import Path
from typing import List

from readmeai.cli.options import BadgeOptions
from readmeai.config.settings import ConfigLoader
from readmeai.generators import badges, tables, tree, utils
from readmeai.generators.quickstart import get_setup_data
from readmeai.services.git import GitHost


class MarkdownBuilder:
    """Builds each section of the README Markdown file."""

    def __init__(
        self,
        config_loader: ConfigLoader,
        dependencies: List[str],
        summaries: tuple,
        temp_dir: str,
    ):
        """Initializes the MarkdownBuilder class."""
        self.config_loader = config_loader
        self.config = config_loader.config
        self.deps = dependencies
        self.summaries = summaries
        self.temp_dir = Path(temp_dir)
        self.md = self.config.md
        self.full_name = self.config.git.full_name
        self.host = self.config.git.host
        self.host_domain = self.config.git.host_domain
        self.repo_name = self.config.git.name
        self.repo_url = self.config.git.repository
        if self.host_domain == GitHost.LOCAL:
            self.repo_url = f"../{self.repo_name}"

    @property
    def md_header(self) -> str:
        """Generates the README header section."""
        if BadgeOptions.SKILLS.value not in self.md.badge_style:
            md_shields, md_badges = badges.shields_icons(
                self.config, self.deps, self.full_name, self.host
            )
        else:
            md_shields = (
                "<!-- Shields.io badges not used with skill icons. -->"
            )
            md_badges = badges.skill_icons(self.config, self.deps)

        return self.md.header.format(
            alignment=self.md.alignment,
            image=self.md.image,
            repo_name=self.repo_name.upper(),
            slogan=self.md.slogan,
            shields_icons=md_shields,
            badge_icons=md_badges,
        )

    @property
    def md_summaries(self) -> str:
        """Generates the README code summaries section."""
        summaries = tables.format_code_summaries(
            self.md.placeholder,
            self.summaries,
        )
        md_summaries = tables.generate_markdown_tables(
            self.md.modules_widget,
            summaries,
            self.full_name,
            self.repo_url,
        )
        return md_summaries

    @property
    def md_tree(self) -> str:
        """Generates the README directory tree structure."""
        md_tree = tree.TreeGenerator(
            repo_name=self.repo_name,
            root_dir=self.temp_dir,
            repo_url=self.repo_url,
            max_depth=self.md.tree_depth,
        ).tree()
        return self.md.tree.format(md_tree)

    @property
    def md_quickstart(self) -> str:
        """Generates the README Getting Started section."""
        setup_data = get_setup_data(self.config_loader, self.summaries)
        return self.md.quickstart.format(
            repo_name=self.repo_name,
            repo_url=self.repo_url,
            install_command=setup_data.install_command,
            run_command=setup_data.run_command,
            test_command=setup_data.test_command,
            system_requirements=setup_data.prerequisites,
        )

    @property
    def md_contributing(self) -> str:
        """Generates the README Contributing section."""
        return self.md.contribute.format(
            host=self.host,
            host_domain=self.host_domain,
            full_name=self.full_name,
            repo_name=self.repo_name,
            repo_url=self.repo_url,
        )

    def build(self) -> str:
        """Builds the README Markdown file."""
        md_contents = [
            self.md_header,
            self.md.toc.format(repo_name=self.repo_name),
            self.md.overview,
            self.md.features,
            self.md_tree,
            self.md.modules,
            self.md_summaries,
            self.md_quickstart,
            self.md_contributing,
        ]

        if self.config.md.emojis is False:
            md_contents = utils.remove_emojis(md_contents)

        return "\n".join(md_contents)



"""Dynamically creates the 'Quickstart' section of the README file."""

import traceback
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List

from readmeai.config.settings import ConfigLoader
from readmeai.core.logger import Logger

_logger = Logger(__name__)


@dataclass
class QuickStart:
    """Information about using, running, and testing a repository."""

    install_command: str
    run_command: str
    test_command: str
    prerequisites: str
    language_counts: Dict[str, int]
    language_key: str
    language_name: str = None


def count_languages(
    summaries: List[str], config_loader: ConfigLoader
) -> Dict[str, int]:
    """
    Counts the occurrences of each language in the summaries.
    """
    parser_files = config_loader.parsers.get("parsers")

    language_counts = {}

    for file_path, _ in summaries:
        language = Path(file_path).suffix[1:]

        if str(file_path) in [
            dependency_file for dependency_file in parser_files
        ]:
            continue

        if language and language not in config_loader.blacklist:
            language_counts[language] = language_counts.get(language, 0) + 1

    return language_counts


def get_top_language(language_counts: Dict[str, int]) -> str:
    """
    Determines the top language.
    """
    if not language_counts:
        return None

    return max(sorted(language_counts), key=language_counts.get)


def get_top_language_setup(
    language_counts: Dict[str, int], config_loader: ConfigLoader
) -> QuickStart:
    """
    Determines the top language and retrieves its setup commands.
    """
    if not language_counts:
        return None

    languages = config_loader.languages.get("language_names")
    commands = config_loader.commands.get("quickstart_guide")

    language_key = get_top_language(language_counts)
    language_name = languages.get(language_key, languages.get("default"))
    quickstart_commands = commands.get(language_name, commands.get("default"))
    prerequisites = f"**{language_name}**: `version x.y.z`"

    return QuickStart(
        *quickstart_commands,
        prerequisites,
        language_counts,
        language_key,
        language_name,
    )


def get_setup_data(
    config_loader: ConfigLoader, summaries: List[str]
) -> QuickStart:
    """
    Generates the 'Quick Start' section of the README file.
    """
    default_setup = QuickStart("", "", "", {}, "", "")

    try:
        language_counts = count_languages(summaries, config_loader)
        setup = (
            get_top_language_setup(language_counts, config_loader)
            or default_setup
        )

    except Exception as exc:
        _logger.debug(f"Exception: {exc}\n{traceback.format_exc()}")
        setup = default_setup

    _logger.info(f"Quickstart information: {setup}")

    return setup



"""
Custom logger implementation for the readme-ai package.
"""

import logging
import sys

from colorlog import ColoredFormatter

LOG_LEVEL_EMOJIS = {
    "DEBUG": "⚙︎",
    "INFO": "►",
    "WARNING": "⚠️",
    "ERROR": "ⓧ",
    "CRITICAL": "‼",
}


class CustomFormatter(ColoredFormatter):
    """Custom colored log formatter."""

    def format(self, record) -> str:
        """Formats the log record."""
        record.emoji = LOG_LEVEL_EMOJIS.get(record.levelname, "")
        return super().format(record)


class Logger:
    """Custom logger implementation."""

    _instances = {}

    def __new__(cls, name, level="DEBUG"):
        if name not in cls._instances:
            instance = super().__new__(cls)
            instance._name = name
            instance._level = level
            instance._configure_logger()
            cls._instances[name] = instance
        return cls._instances[name]

    def _configure_logger(self):
        """Configures the logger."""
        formatter = CustomFormatter(
            "%(log_color)s%(emoji)s %(levelname)s | %(white)s%(asctime)s | %(cyan)s%(name)s | %(purple)s%(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            reset=True,
            log_colors={
                "DEBUG": "blue",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red,bg_white",
            },
        )
        handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(formatter)
        logger = logging.getLogger(self._name)
        logger.addHandler(handler)
        logger.setLevel(self._level)

    def info(self, msg, *args, **kwargs):
        """Logs an info message."""
        logging.getLogger(self._name).info(msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        """Logs a debug message."""
        logging.getLogger(self._name).debug(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """Logs a warning message."""
        logging.getLogger(self._name).warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """Logs an error message."""
        logging.getLogger(self._name).error(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """Logs a critical message."""
        logging.getLogger(self._name).critical(msg, *args, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        """Logs a message at the specified level."""
        logging.getLogger(self._name).log(level, msg, *args, **kwargs)



"""
Model factory that returns the appropriate LLM handler based on CLI input.
"""

from readmeai._exceptions import UnsupportedServiceError
from readmeai.cli.options import ModelOptions as llms
from readmeai.config.settings import ConfigLoader
from readmeai.core.models import BaseModelHandler
from readmeai.models.offline import OfflineHandler
from readmeai.models.openai import OpenAIHandler
from readmeai.models.vertex import VertexAIHandler

MODEL_REGISTRY: dict[str, BaseModelHandler] = {
    llms.OFFLINE.name: OfflineHandler,
    llms.OLLAMA.name: OpenAIHandler,
    llms.OPENAI.name: OpenAIHandler,
    llms.VERTEX.name: VertexAIHandler,
}


def model_handler(conf: ConfigLoader) -> BaseModelHandler:
    """Registry that returns the appropriate LLM handler based on CLI input."""
    llm_handler = MODEL_REGISTRY.get(conf.config.llm.api)

    if llm_handler is None:
        raise UnsupportedServiceError(
            f"Unsupported LLM service provided: {conf.config.llm.api}"
        )

    return llm_handler(conf)



"""
CLI entrypoint for the readme-ai package.
"""

from typing import Optional

import click

from readmeai._agent import readme_agent
from readmeai.cli import options


@click.command()
@options.alignment
@options.api
@options.badge_color
@options.badge_style
@options.base_url
@options.context_window
@options.emojis
@options.image
@options.language
@options.model
@options.output
@options.rate_limit
@options.repository
@options.temperature
@options.tree_depth
@options.template
@options.top_p
def main(
    alignment: Optional[str],
    api: Optional[str],
    badge_color: Optional[str],
    badge_style: Optional[str],
    base_url: Optional[str],
    context_window: Optional[int],
    emojis: Optional[bool],
    image: Optional[str],
    language: Optional[str],
    model: Optional[str],
    output: Optional[str],
    rate_limit: Optional[int],
    repository: str,
    temperature: Optional[float],
    template: Optional[str],
    top_p: Optional[float],
    tree_depth: Optional[int],
) -> None:
    """Entry point for the readme-ai CLI application."""
    readme_agent(
        alignment=alignment,
        api=api,
        badge_color=badge_color,
        badge_style=badge_style,
        base_url=base_url,
        context_window=context_window,
        emojis=emojis,
        image=image,
        # language=language,
        model=model,
        output_file=output,
        rate_limit=rate_limit,
        repository=repository,
        temperature=temperature,
        # template=template,
        top_p=top_p,
        tree_depth=tree_depth,
    )


if __name__ == "__main__":
    main()



"""Dependency file parsers for C/C++ projects."""

import re
from typing import List

from readmeai.core.parsers import BaseFileParser


class CMakeParser(BaseFileParser):
    """Parser for CMake dependency files."""

    def __init__(self) -> None:
        """Initializes the handler with given configuration."""
        super().__init__()

    def parse(self, content: str) -> List[str]:
        """Extracts dependencies, libs, and software from a CMakeLists.txt."""
        try:
            extracted_dependencies = []
            for line in content.splitlines():
                line = line.strip()
                if (
                    line.startswith("find_package")
                    or "target_link_libraries" in line
                ):
                    dependencies = re.findall(
                        r"(?:find_package\(|target_link_libraries\()\s*(\w+)",
                        line,
                    )
                    extracted_dependencies.extend(dependencies)

                if line.startswith("target_link_libraries") or line.startswith(
                    "find_package"
                ):
                    libs = re.findall(
                        r"target_link_libraries\([^)]+\s+([^)]+)\)", line
                    )
                    extracted_dependencies.extend(libs)

            return list(set(extracted_dependencies))

        except re.error as exc:
            return self.handle_parsing_error(f"CMakeLists.txt: {str(exc)}")


class ConfigureAcParser(BaseFileParser):
    """Parser for configure.ac dependency files."""

    def parse(self, content: str) -> List[str]:
        """Extracts package names from a configure.ac file."""
        try:
            regex = re.compile(r"AC_CHECK_LIB\([^)]+\s+([^)]+)\)")
            return regex.findall(content)
        except re.error as exc:
            return self.handle_parsing_error(f"configure.ac: {str(exc)}")


class MakefileAmParser(BaseFileParser):
    """Parser for Makefile dependency files."""

    def parse(self, content: str) -> List[str]:
        """Extracts dependencies from Makefile.am files."""
        try:
            extracted_packages = []
            bin_programs_regex = r"bin_PROGRAMS\s*=\s*(.*)"
            match = re.search(bin_programs_regex, content)

            if match:
                programs = match.group(1).split()
                extracted_packages.extend(programs)

            lib_sources_regex = r"libfoo_la_SOURCES\s*=\s*(.*)"
            match = re.search(lib_sources_regex, content)

            if match:
                sources = match.group(1).split()
                extracted_packages.extend(sources)

            check_programs_regex = r"check_PROGRAMS\s*=\s*(.*)"
            match = re.search(check_programs_regex, content)

            if match:
                programs = match.group(1).split()
                extracted_packages.extend(programs)

            check_libs_regex = r"check_LTLIBRARIES\s*=\s*(.*)"
            match = re.search(check_libs_regex, content)

            return list(set(extracted_packages))

        except re.error as exc:
            return self.handle_parsing_error(f"Makefile.am: {str(exc)}")



"""Parser for .properties configuration files."""

import re
from typing import List

from readmeai.core.parsers import BaseFileParser


class PropertiesParser(BaseFileParser):
    """Parser for .properties files."""

    def parse(self, content: str) -> List[str]:
        """Parse the content of a .properties file."""
        names = []

        # Extracting jdbc connection strings
        regex = r"jdbc:(\w+)://"
        match = re.search(regex, content)
        if match:
            names.append(match.group(1))

        # Other packages
        regex = r"(?P<name>[\w.]+)[\s]*="
        for match in re.finditer(regex, content):
            names.append(match.group("name"))

        return names



"""
Abstract base class for the large language model (LLM) API handlers.
"""

import asyncio
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from typing import Any, Dict, Generator, List, Tuple, Union

import aiohttp

from readmeai.cli.options import ModelOptions as llms
from readmeai.config.settings import ConfigLoader
from readmeai.core.logger import Logger
from readmeai.models.prompts import (
    get_prompt_context,
    set_additional_contexts,
    set_summary_context,
)
from readmeai.models.tokens import update_max_tokens


class BaseModelHandler(ABC):
    """Abstract base class for Large Language Model (LLM) API handlers."""

    def __init__(self, config_loader: ConfigLoader) -> None:
        """Initializes the LLM handler with configuration and logging."""
        self._logger = Logger(__name__)
        self._session: aiohttp.ClientSession = None
        self.config = config_loader.config
        self.prompts = config_loader.prompts
        self.sys_content = self.config.api.content
        self.rate_limit = self.config.api.rate_limit
        self.rate_limit_semaphore = asyncio.Semaphore(self.rate_limit)

    @abstractmethod
    async def _model_settings(self):
        """Initializes the LLM settings for the specific API implementation."""
        ...

    @abstractmethod
    async def _build_payload(self, prompt: str, tokens: int) -> Dict[str, Any]:
        """Builds the payload for the POST request to the LLM API."""
        ...

    @abstractmethod
    async def _make_request(
        self,
        index: str,
        prompt: str,
        tokens: int,
        raw_files: List[Tuple[str, str]] = None,
    ) -> Tuple[str, str]:
        """Handles LLM API response and returns the generated text."""
        ...

    @asynccontextmanager
    async def use_api(self) -> Generator:
        """Async context manager for managing the lifecycle of the HTTP client."""
        async with aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(total=None)
        ) as session:
            self._session = session
            try:
                yield self
            finally:
                await self.close()

    async def close(self) -> None:
        """Closes the HTTP client session."""
        if self._session:
            await self._session.close()
            self._session = None
        else:
            self._logger.debug("HTTP client already closed.")

    async def batch_request(
        self,
        dependencies: List[str],
        raw_files: List[Tuple[str, str]],
    ) -> List[str]:
        """Generates a batch of prompts and processes the responses.

        Parameters
        ----------
        dependencies
            List of dependencies for the project.
        raw_files
            List of tuples containing all file paths and file contents.

        Returns
        -------
            List of generated text responses from the LLM API.
        """
        if self.config.llm.api == llms.OFFLINE.name:
            return await self._make_request(
                index=None, prompt=None, tokens=None, raw_files=raw_files
            )
        else:
            raw_files = [
                file for file in raw_files if not file[0].endswith(".lock")
            ]
        file_summary_prompts = await set_summary_context(
            self.config, dependencies, raw_files
        )
        file_summary_responses = await self._batch_prompts(
            file_summary_prompts
        )
        additional_prompts = await set_additional_contexts(
            self.config, dependencies, file_summary_responses
        )
        additional_responses = await self._batch_prompts(additional_prompts)

        return file_summary_responses + additional_responses

    async def _batch_prompts(
        self, prompts: List[Union[str, Tuple[str, str]]], batch_size: int = 10
    ) -> List[str]:
        """Processes a batch of prompts and returns the generated text.

        Parameters
        ----------
        prompts
            List of prompts to be processed.
        batch_size, optional
            Max number of prompts to process in a single batch. Defaults to 10.

        Returns
        -------
            List of generated text responses from the LLM API.
        """
        responses = []

        for batch in self._generate_batches(prompts, batch_size):
            batch_responses = await asyncio.gather(
                *[self._process_batch(prompt) for prompt in batch],
                return_exceptions=True,
            )
            responses.extend(batch_responses)

        return responses

    def _generate_batches(
        self, items: List[Any], batch_size: int
    ) -> Generator[List[Any], None, None]:
        """Generates batches of items to be processed.

        Parameters
        ----------
        items
            List of items to be processed.
        batch_size
            Maximum number of items to process in a single batch.

        Returns
        -------
            List of batches of items to be processed.

        Yields
        ------
            List of items to be processed.
        """
        for i in range(0, len(items), batch_size):
            yield items[i : i + batch_size]

    async def _process_batch(self, prompt: Dict[str, Any]) -> str:
        """Processes a single prompt and returns the generated text.

        Parameters
        ----------
        prompt
            Prompt to be processed by the LLM API.

        Returns
        -------
            Generated text response from the LLM API.
        """
        if prompt["type"] == "file_summary":
            return await self._make_request_code_summary(prompt["context"])
        else:
            formatted_prompt = get_prompt_context(
                self.prompts, prompt["type"], prompt["context"]
            )
            tokens = update_max_tokens(self.tokens, formatted_prompt)
            _, summary = await self._make_request(
                prompt["type"], formatted_prompt, tokens
            )
            return summary

    async def _make_request_code_summary(
        self, file_context: List[Tuple[str, str]]
    ) -> List[Tuple[str, str]]:
        """Generates code summaries for each file in the project.

        Parameters
        ----------
        file_context
            List of tuples containing all file paths and file contents.

        Returns
        -------
            List of generated code summaries for each file in the project.
        """
        summary_text = []

        for file_path, file_content in file_context["file_summary"]:
            prompt = self.prompts["prompts"]["file_summary"].format(
                self.config.md.tree, file_path, file_content
            )
            tokens = update_max_tokens(self.tokens, prompt)
            _, summary_or_error = await self._make_request(
                file_path, prompt, tokens
            )
            summary_text.append((file_path, summary_or_error))

        return summary_text



"""
OpenAI API LLM handler implementation, with Ollama support.
"""

import os
from typing import List, Tuple

import aiohttp
import openai
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from readmeai.cli.options import ModelOptions as llms
from readmeai.config.settings import ConfigLoader
from readmeai.core.models import BaseModelHandler
from readmeai.models.tokens import token_handler
from readmeai.utils.text_cleaner import clean_response

_localhost = "http://localhost:11434/v1/"


class OpenAIHandler(BaseModelHandler):
    """OpenAI API LLM implementation."""

    def __init__(self, config_loader: ConfigLoader) -> None:
        """Initialize OpenAI API LLM handler."""
        super().__init__(config_loader)
        self._model_settings()

    def _model_settings(self):
        """Set default values for OpenAI API."""
        self.model = self.config.llm.model
        self.temperature = self.config.llm.temperature
        self.tokens = self.config.llm.tokens
        self.top_p = self.config.llm.top_p

        if self.config.llm.api == llms.OPENAI.name:
            self.endpoint = self.config.llm.base_url
            self.client = openai.OpenAI(
                api_key=os.environ.get("OPENAI_API_KEY")
            )
        elif self.config.llm.api == llms.OLLAMA.name:
            self.endpoint = f"{_localhost}chat/completions"
            self.client = openai.OpenAI(
                base_url=_localhost, api_key=llms.OLLAMA.name
            )

        self.headers = {"Authorization": f"Bearer {self.client.api_key}"}

    async def _build_payload(self, prompt: str, tokens: int) -> dict:
        """Build payload for POST request to OpenAI API."""
        return {
            "messages": [
                {
                    "role": "system",
                    "content": self.sys_content,
                },
                {"role": "user", "content": prompt},
            ],
            "model": self.model,
            "max_tokens": tokens,
            "temperature": self.temperature,
        }

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=4, max=10),
        retry=retry_if_exception_type(
            (
                aiohttp.ClientError,
                aiohttp.ClientResponseError,
                aiohttp.ClientConnectorError,
                openai.OpenAIError,
            )
        ),
    )
    async def _make_request(
        self,
        index: str,
        prompt: str,
        tokens: int,
        raw_files: List[Tuple[str, str]] = None,
    ) -> Tuple[str, str]:
        """Processes OpenAI API LLM responses and returns generated text."""
        try:
            prompt = await token_handler(self.config, index, prompt, tokens)

            data = await self._build_payload(prompt, tokens)

            async with self._session.post(
                self.endpoint,
                headers=self.headers,
                json=data,
            ) as response:
                response.raise_for_status()
                response = await response.json()
                text = response["choices"][0]["message"]["content"]
                self._logger.info(f"Response for '{index}':\n{text}")
                return index, clean_response(index, text)

        except (
            aiohttp.ClientError,
            aiohttp.ClientResponseError,
            aiohttp.ClientConnectorError,
            openai.OpenAIError,
        ) as exc:
            message = f"Error making request for - `{index}`: {exc}"
            self._logger.error(message)
            return index, self.config.md.placeholder



"""Parser utilities for Java-based dependency files."""

import re
from typing import List

from readmeai.core.parsers import BaseFileParser


class MavenParser(BaseFileParser):
    """Parser for Maven dependency files in pom.xml format."""

    def __init__(self) -> None:
        """Initializes the handler with given configuration."""
        super().__init__()

    def parse(self, content: str) -> List[str]:
        """Extract packages names from Maven pom.xml files."""
        try:
            regex = re.compile(
                r"<dependency>\s*<groupId>([^<]+)</groupId>\s*<artifactId>([^<]+)</artifactId>\s*<version>([^<]+)</version>"
            )
            matches = regex.findall(content)
            dependencies = [
                artifact_id for group_id, artifact_id, version in matches
            ]
            if any("spring" in dependency for dependency in dependencies):
                dependencies.append("spring")
            return set(dependencies)

        except re.error as exc:
            return self.handle_parsing_error(f"pom.xml: {str(exc)}")



"""Creates markdown tables to store LLM text responses in the README file."""

from pathlib import Path
from typing import List, Tuple

from readmeai.core.logger import Logger
from readmeai.services.git import fetch_git_file_url

_logger = Logger(__name__)


def construct_markdown_table(
    data: List[Tuple[str, str]], repo_url: str, full_name: str
) -> str:
    """Builds a Markdown table from the provided data."""
    headers = ["File", "Summary"]
    table_rows = [headers, ["---", "---"]]

    for module, summary in data:
        file_name = str(Path(module).name)
        if "invalid" in full_name.lower():
            return file_name
        host_url = fetch_git_file_url(module, full_name, repo_url)
        md_format_host_url = f"[{file_name}]({host_url})"
        table_rows.append([md_format_host_url, summary])

    return format_as_markdown_table(table_rows)


def extract_folder_name(module: str) -> str:
    """Extracts the folder name from a module path."""
    path_parts = Path(module).parts
    return ".".join(path_parts[:-1]) if len(path_parts) > 1 else "."


def format_as_markdown_table(rows: List[List[str]]) -> str:
    """Formats rows of data as a Markdown table."""
    max_column_widths = [
        max(len(str(row[col])) for row in rows) for col in range(len(rows[0]))
    ]

    formatted_lines = [
        "| "
        + " | ".join(
            str(item).ljust(width)
            for item, width in zip(row, max_column_widths)
        )
        + " |"
        for row in rows
    ]

    return "\n".join(formatted_lines)


def format_code_summaries(
    placeholder: str, code_summaries: Tuple[str, str]
) -> List[Tuple[str, str]]:
    """Converts the given code summaries into a formatted list."""
    _logger.debug(f"Formatting code summaries: {code_summaries}")

    formatted_summaries = []

    for summary in code_summaries:
        if is_valid_tuple_summary(summary):
            module, summary_text = summary
        else:
            module, summary_text = summary, placeholder

        formatted_summaries.append((module, summary_text))

    return formatted_summaries


def generate_markdown_tables(
    table_widget: str,
    summaries: List[Tuple[str, str]],
    project_name: str,
    repository_url: str,
) -> str:
    """Produces Markdown tables for each project sub-directory."""
    summaries_by_folder = group_summaries_by_folder(summaries)

    markdown_tables = []
    for folder, entries in summaries_by_folder.items():
        table_in_markdown = construct_markdown_table(
            entries, repository_url, project_name
        )
        table_wrapper = table_widget.format(folder, table_in_markdown)
        markdown_tables.append(table_wrapper)

    return "\n".join(markdown_tables)


def group_summaries_by_folder(summaries: List[Tuple[str, str]]) -> dict:
    """Groups code summaries by their sub-directory."""
    folder_map = {}
    for module, summary in summaries:
        folder_name = extract_folder_name(module)
        folder_map.setdefault(folder_name, []).append((module, summary))
    return folder_map


def is_valid_tuple_summary(summary: Tuple[str, str]) -> bool:
    """Checks if a summary is a valid tuple format."""
    return isinstance(summary, tuple) and len(summary) == 2



"""Git host providers helper methods to retrieve repository metadata."""

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import aiohttp

from readmeai.core.logger import Logger
from readmeai.services.git import fetch_git_api_url

_logger = Logger(__name__)


@dataclass
class RepositoryMetadata:
    """Dataclass to store GitHub repository metadata."""

    name: str
    full_name: str
    owner: str
    owner_url: Optional[str]
    description: Optional[str]

    # Repository statistics
    stars_count: int
    forks_count: int
    watchers_count: int
    open_issues_count: int

    # Repository details
    default_branch: str
    created_at: str
    updated_at: str
    pushed_at: str
    size_kb: int

    # Repository URLs
    clone_url_http: str
    clone_url_ssh: str
    contributors_url: Optional[str]
    languages_url: str
    issues_url: Optional[str]

    # Programming languages and topics
    language: Optional[str]
    languages: List[str]
    topics: List[str]

    # Additional repository settings
    has_wiki: bool
    has_issues: bool
    has_projects: bool
    is_private: bool
    homepage_url: Optional[str]

    # License information
    license_name: Optional[str]
    license_url: Optional[str]


def _parse_repository_metadata(repo_data: dict) -> RepositoryMetadata:
    """Converts raw repository data from GitHub API into dataclass."""
    languages = repo_data.get("languages", {})
    license_info = repo_data.get("license", {}) or {}
    owner_info = repo_data.get("owner", {}) or {}

    return RepositoryMetadata(
        name=repo_data.get("name", ""),
        full_name=repo_data.get("full_name", ""),
        owner=owner_info.get("login", ""),
        owner_url=owner_info.get("html_url", ""),
        description=repo_data.get("description", ""),
        stars_count=repo_data.get("stargazers_count", 0),
        forks_count=repo_data.get("forks_count", 0),
        watchers_count=repo_data.get("watchers_count", 0),
        open_issues_count=repo_data.get("open_issues_count", 0),
        default_branch=repo_data.get("default_branch", ""),
        created_at=repo_data.get("created_at", ""),
        updated_at=repo_data.get("updated_at", ""),
        pushed_at=repo_data.get("pushed_at", ""),
        size_kb=repo_data.get("size", 0),
        clone_url_http=repo_data.get("clone_url", ""),
        clone_url_ssh=repo_data.get("ssh_url", ""),
        contributors_url=repo_data.get("contributors_url"),
        languages_url=repo_data.get("languages_url", ""),
        issues_url=repo_data.get("issues_url"),
        language=repo_data.get("language", ""),
        languages=list(languages.keys()) if languages else [],
        topics=repo_data.get("topics", []),
        has_wiki=repo_data.get("has_wiki", False),
        has_issues=repo_data.get("has_issues", False),
        has_projects=repo_data.get("has_projects", False),
        is_private=repo_data.get("private", False),
        homepage_url=repo_data.get("homepage", ""),
        license_name=license_info.get("name", ""),
        license_url=license_info.get("url", ""),
    )


async def _fetch_repository_metadata(
    session: aiohttp.ClientSession, url: str, **kwargs
) -> Dict[str, Any]:
    """Fetches repository metadata from the git host provider."""
    async with session.get(url, **kwargs) as response:
        response.raise_for_status()
        if response.status != 200:
            raise aiohttp.ClientResponseError(
                request_info=response.request_info,
                history=response.history,
                status=response.status,
            )
        return await response.json()


async def fetch_git_repository_metadata(
    session: aiohttp.ClientSession, repository: str
) -> Optional[RepositoryMetadata]:
    """Retrieves GitHub repository metadata and returns a dataclass."""
    api_url = await fetch_git_api_url(repository)
    if not api_url:
        return None

    try:
        metadata = await _fetch_repository_metadata(session, api_url)
        return _parse_repository_metadata(metadata) if metadata else None

    except aiohttp.ClientError as exc:
        _logger.error(
            f"Client error while fetching repository metadata: {exc}"
        )
        return None



"""Parser for Docker (Dockerfile, docker-compose.yaml) configuration files."""

import re
from typing import List

import yaml

from readmeai.core.parsers import BaseFileParser


class DockerfileParser(BaseFileParser):
    """Parser for Dockerfile dependency files."""

    def __init__(self) -> None:
        """Initializes the handler with given configuration."""
        super().__init__()

    def parse(self, content: str) -> List[str]:
        """Extracts package names from a Dockerfile."""
        try:
            dependencies = []
            lines = content.split("\n")
            for line in lines:
                if line.startswith("FROM"):
                    match = re.search(
                        r"FROM\s+(?:--platform=[^\s]+\s+)?([^\s:]+):?([^\s]*)",
                        line,
                    )
                    if match:
                        base_image, version = match.groups()
                        if not version:
                            version = "latest"
                        dependencies.append((base_image, version))

            return dependencies

        except re.error as exc:
            return self.handle_parsing_error(f"Dockerfile: {str(exc)}")


class DockerComposeParser(BaseFileParser):
    """Parser for Docker related files."""

    def parse(self, content: str) -> List[str]:
        """Parse docker-compose.yaml file and return a list of services."""
        try:
            data = yaml.safe_load(content)
            if isinstance(data, dict) and "services" in data:
                return list(data["services"].keys())

        except yaml.YAMLError as exc:
            return self.handle_parsing_error(f"Dockerfile: {str(exc)}")



"""
Model handler for running the CLI without a LLM API service.
"""

from typing import Any, Dict, List, Tuple

from readmeai.config.settings import ConfigLoader
from readmeai.core.models import BaseModelHandler


class OfflineHandler(BaseModelHandler):
    """Offline model handler, used when no LLM API is provided."""

    def __init__(self, config_loader: ConfigLoader) -> None:
        """Initialize the OfflineHandler class."""
        super().__init__(config_loader)
        self._model_settings()

    def _model_settings(self):
        """Set default values for offline mode."""
        self.placeholder = self.config.md.placeholder

    async def _build_payload(self, prompt: str, tokens: int) -> Dict[str, Any]:
        """Builds the payload for the POST request to the LLM API."""
        ...

    async def _make_request(
        self,
        index: str = None,
        prompt: str = None,
        tokens: int = None,
        raw_files: List[Tuple[str, str]] = None,
    ) -> Tuple[str, str]:
        """Returns placeholder text where LLM API response would be."""
        file_summaries = [
            (str(file_path), self.placeholder) for file_path, _ in raw_files
        ]
        return (
            file_summaries,
            self.placeholder,
            self.placeholder,
            self.placeholder,
        )



"""Abstract factory module for all project file parsers."""

from typing import Dict, Type

from readmeai.core.parsers import BaseFileParser
from readmeai.parsers.configuration.docker import (
    DockerComposeParser,
    DockerfileParser,
)
from readmeai.parsers.configuration.properties import PropertiesParser
from readmeai.parsers.language.cpp import (
    CMakeParser,
    ConfigureAcParser,
    MakefileAmParser,
)
from readmeai.parsers.language.go import GoModParser
from readmeai.parsers.language.python import (
    RequirementsParser,
    TomlParser,
    YamlParser,
)
from readmeai.parsers.language.rust import CargoTomlParser
from readmeai.parsers.language.swift import SwiftPackageParser
from readmeai.parsers.package.gradle import (
    BuildGradleKtsParser,
    BuildGradleParser,
)
from readmeai.parsers.package.maven import MavenParser
from readmeai.parsers.package.npm import PackageJsonParser
from readmeai.parsers.package.yarn import YarnLockParser

ParserRegistryType = dict[str, Type[BaseFileParser]]

PARSER_REGISTRY = {
    # Configuration
    ".properties": PropertiesParser,
    # Language/Framework
    # Python
    "Pipfile": TomlParser(),
    "pyproject.toml": TomlParser(),
    "requirements.in": RequirementsParser(),
    "requirements.txt": RequirementsParser(),
    "requirements-dev.txt": RequirementsParser(),
    "requirements-test.txt": RequirementsParser(),
    "requirements-prod.txt": RequirementsParser(),
    "dev-requirements.txt": RequirementsParser(),
    "environment.yml": YamlParser(),
    "environment.yaml": YamlParser(),
    # "setup.py": setup_py_parser,
    # "setup.cfg": setup_cfg_parser,
    # C/C++
    "cmakeLists.txt": CMakeParser(),
    "configure.ac": ConfigureAcParser(),
    "Makefile.am": MakefileAmParser(),
    # JavaScript/Node.js
    "package.json": PackageJsonParser(),
    "yarn.lock": YarnLockParser(),
    # Kotlin and Kotlin DSL
    "build.gradle": BuildGradleParser(),
    "build.gradle.kts": BuildGradleKtsParser(),
    # Go
    "go.mod": GoModParser(),
    # Java
    "pom.xml": MavenParser(),
    # Rust
    "cargo.toml": CargoTomlParser(),
    # Swift
    "Package.swift": SwiftPackageParser(),
    "Dockerfile": DockerfileParser(),
    "docker-compose.yaml": DockerComposeParser(),
    # Package Managers
    # Monitoring and Logging
}


def parser_handler() -> Dict[str, BaseFileParser]:
    """Returns a dictionary of callable file parser methods."""
    return PARSER_REGISTRY



"""
Pre-processes the input repository files and extracts metadata.
"""

from dataclasses import dataclass, field
from pathlib import Path
from typing import Generator, List, Tuple

from readmeai.cli.options import ModelOptions as Models
from readmeai.config.settings import ConfigLoader
from readmeai.core.logger import Logger
from readmeai.generators.builder import MarkdownBuilder
from readmeai.models.tokens import count_tokens
from readmeai.parsers.factory import parser_handler

_github_actions_path = ".github/workflows"


@dataclass
class FileContext:
    """Data class to store file contents and metadata."""

    file_path: Path
    file_name: str
    file_ext: str
    content: str
    tokens: int = 0
    language: str = field(init=False)
    dependencies: List[str] = field(default_factory=list)

    def __post_init__(self):
        """Initializes the FileContext class."""
        self.file_ext = (
            self.file_name.split(".")[-1] if "." in self.file_name else ""
        )
        self.language = self.file_ext.lower()


class RepositoryProcessor:
    """Processes the repository files and generates FileContext list"""

    def __init__(self, config_loader: ConfigLoader):
        """Initializes the RepositoryProcessor class."""
        self._logger = Logger(__name__)
        self.config_loader = config_loader
        self.config = config_loader.config
        self.blacklist = config_loader.blacklist.get("blacklist")
        self.commands = config_loader.commands
        self.languages = config_loader.languages.get("language_names")
        self.parser_files = config_loader.parsers.get("parsers")

    def create_file_data(
        self, file_info: Tuple[str, Path, str]
    ) -> FileContext:
        """Creates a FileContext instance from the file information."""
        file_name, file_path, content = file_info
        return FileContext(
            file_name=file_name,
            file_path=file_path,
            content=content,
            file_ext="",
        )

    def extract_dependencies(self, file_data: FileContext) -> List[str]:
        """Extracts the dependency file contents using the factory pattern."""
        parsers = parser_handler()

        if file_data.file_name not in parsers:
            return []

        parser = parsers.get(file_data.file_name)
        dependency_names = parser.parse(content=file_data.content)

        self._logger.info(
            f"Dependency file found: {file_data.file_name}:\n{dependency_names}"
        )

        return dependency_names

    def generate_contents(self, repo_path: str) -> List[FileContext]:
        """Generates a List of Dict of file information."""
        if isinstance(repo_path, str):
            repo_path = Path(repo_path)

        return [file_data for file_data in self.generate_file_info(repo_path)]

    def generate_file_info(
        self, repo_path: Path
    ) -> Generator[FileContext, None, None]:
        """
        Generates FileContext instances for each file in the repository.
        """
        for file_path in repo_path.rglob("*"):
            if self._filter_file(file_path):
                continue
            file_data = self._process_file_path(file_path, repo_path)
            if file_data:
                yield file_data

    def get_dependencies(self, contents: List[FileContext]) -> List[str]:
        """Returns a list of dependencies."""
        try:
            dependency_dict = {}
            dependencies = set()
            parser_files = self.config_loader.parsers.get("parsers")

            for file_data in contents:
                dependencies.update(file_data.dependencies)
                dependencies.add(file_data.language.lower())
                dependencies.add(file_data.file_ext)

                if file_data.file_name in parser_files["parsers"]:
                    dependencies.add(file_data.file_name)
                    dependency_dict[
                        file_data.file_name
                    ] = file_data.dependencies

                if _github_actions_path in str(file_data.file_path):
                    dependencies.add("github actions")

            return list(dependencies), dependency_dict

        except Exception as exc:
            self._logger.error(f"Error getting dependencies: {exc}")
            return [], {}

    def _filter_file(self, file_path: Path) -> bool:
        """
        Determines if a file should be ignored based on configurations.
        """
        if (
            is_file_ignored(self.config_loader, file_path)
            and str(file_path.name) in self.parser_files
        ):
            return False

        return not file_path.is_file() or is_file_ignored(
            self.config_loader, file_path
        )

    def _process_file_path(
        self, file_path: Path, repo_path: Path
    ) -> FileContext:
        """
        Processes an individual file path and returns FileContext.
        """
        relative_path = file_path.relative_to(repo_path)
        if _github_actions_path in str(relative_path):
            return FileContext(
                file_path=relative_path,
                file_name="github actions",
                file_ext="",
                content="",
            )

        try:
            with file_path.open(encoding="utf-8") as file:
                content = file.read()

            file_data = FileContext(
                file_path=relative_path,
                file_name=file_path.name,
                file_ext=file_path.suffix,
                content=content,
            )

            file_data.dependencies = self.extract_dependencies(file_data)

            try:
                file_data.language = self.languages.get(
                    file_data.file_ext, self.languages.get("default")
                ).lower()
            except Exception:
                file_data.language = None

            return file_data

        except (OSError, UnicodeDecodeError) as exc:
            self._logger.warning(f"Error reading file {file_path}: {exc}")

    def language_mapper(
        self, contents: List[FileContext]
    ) -> List[FileContext]:
        """Maps file extensions to their programming languages."""
        for content in contents:
            content.language = self.languages.get(content.file_ext, "").lower()
        return contents

    def tokenize_content(
        self, contents: List[FileContext]
    ) -> List[FileContext]:
        """Tokenize each file content and return the token count."""
        if self.config.llm.api == Models.OFFLINE.name:
            return contents

        for content in contents:
            content.tokens = count_tokens(
                content.content, self.config.llm.encoder
            )
        return contents


def preprocessor(
    config_loader: ConfigLoader, temp_dir: str
) -> Tuple[List[FileContext], List[str], List[Tuple[str, str]], str]:
    """Processes the repository files and returns the context."""
    config = config_loader.config
    repo_processor = RepositoryProcessor(config_loader)
    repo_context = repo_processor.generate_contents(temp_dir)
    repo_context = repo_processor.language_mapper(repo_context)
    # repo_context = repo_processor.tokenize_content(repo_context)

    dependencies, dependency_dict = repo_processor.get_dependencies(
        repo_context
    )

    raw_files = [
        (str(context.file_path), context.content) for context in repo_context
    ]

    config.md.tree = MarkdownBuilder(
        config_loader, dependencies, raw_files, temp_dir
    ).md_tree

    return dependencies, raw_files


def is_file_ignored(config: ConfigLoader, file_path: Path) -> bool:
    """Determines if a file should be ignored based on the configuration."""
    blacklist = config.blacklist["blacklist"]
    return any(
        [
            file_path.name in blacklist["files"],
            file_path.suffix.lstrip(".") in blacklist["extensions"],
            any(dir in file_path.parts for dir in blacklist["directories"]),
        ]
    )



"""
Custom exceptions for the readme-ai package.
"""

from __future__ import annotations


class ReadmeAIError(Exception):
    """Base class for exceptions in this module."""

    ...


class CLIError(ReadmeAIError):
    """Exceptions related to the CLI."""

    def __init__(self, message, *args):
        super().__init__(f"Invalid option provided to CLI: {message}", *args)


class GitCloneError(ReadmeAIError):
    """Could not clone repository."""

    def __init__(self, repository: str, *args):
        self.repository = repository
        super().__init__(f"Failed to clone repository: {repository}", *args)


class GitValidationError(ReadmeAIError):
    """Could not validate repository."""

    def __init__(self, repository: str, *args):
        self.repository = repository
        super().__init__(f"Failed to validate repository: {repository}", *args)


class FileSystemError(ReadmeAIError):
    """Exceptions related to file system operations."""

    def __init__(self, message, path, *args):
        self.file_path = path
        super().__init__(f"{message}: {path}", *args)


class FileReadError(FileSystemError):
    """Could not read file."""

    ...


class FileWriteError(FileSystemError):
    """Could not write file."""

    ...


class ReadmeGeneratorError(ReadmeAIError):
    """Exceptions related to readme generation."""

    def __init__(self, traceback, *args):
        self.traceback = traceback
        super().__init__(f"Error generating readme: {traceback}", *args)


class UnsupportedServiceError(ReadmeAIError):
    """Exceptions related to the LLMHandler class."""

    def __init__(self, message, *args):
        super().__init__(message, *args)



"""Utility methods to read package resources."""

from importlib import resources
from pathlib import Path

from readmeai._exceptions import FileReadError
from readmeai.core.logger import Logger
from readmeai.utils.file_handler import FileHandler

_logger = Logger(__name__)


def get_resource_path(
    handler: FileHandler,
    file_path: str,
    package: str = "readmeai.config",
    submodule: str = "settings",
) -> dict:
    """Get configuration dictionary from TOML file."""
    try:
        resource_path = resources.files(package).joinpath(submodule, file_path)
        _logger.debug(
            f"Loading file via importlib: {package}.{submodule}.{file_path}"
        )

    except TypeError as exc:
        _logger.debug(f"Error using importlib.resources: {exc}")

        try:
            import pkg_resources

            submodule = submodule.replace(".", "/")
            resource_path = Path(
                pkg_resources.resource_filename(
                    "readmeai", f"{submodule}/{file_path}"
                )
            ).resolve()

        except Exception as exc:
            _logger.error(f"Error loading file via pkg_resources: {exc}")

            raise FileReadError(
                "Error loading file via pkg_resources", str(file_path)
            ) from exc

    if not resource_path.exists():
        _logger.error(f"File not found: {str(resource_path)}")
        raise FileReadError("File not found", str(resource_path))

    if str(file_path).endswith(".toml"):
        return handler.read_toml(str(resource_path))

    elif str(file_path).endswith(".json"):
        return handler.read_json(str(resource_path))

    _logger.error(f"Unsupported file format: {file_path}")

    raise FileReadError("Unsupported file format", str(file_path))



#!/usr/bin/env python3

"""
Runs the README.md file generation process.
"""

__package__ = "readmeai"

import asyncio
import tempfile
import traceback
from pathlib import Path
from typing import Optional

from readmeai._exceptions import ReadmeGeneratorError
from readmeai.cli.options import ImageOptions, prompt_for_image
from readmeai.config.settings import ConfigLoader, GitSettings
from readmeai.core.logger import Logger
from readmeai.core.preprocess import preprocessor
from readmeai.core.utils import get_environment
from readmeai.generators.builder import MarkdownBuilder
from readmeai.models.factory import model_handler
from readmeai.services.git import GitHost, clone_repository
from readmeai.utils.file_handler import FileHandler

_logger = Logger(__name__)


def readme_agent(
    alignment: Optional[str],
    api: Optional[str],
    badge_color: Optional[str],
    badge_style: Optional[str],
    base_url: Optional[str],
    context_window: Optional[int],
    emojis: Optional[bool],
    image: Optional[str],
    # language: Optional[str],
    model: Optional[str],
    output_file: Optional[str],
    rate_limit: Optional[int],
    repository: str,
    temperature: Optional[float],
    # template: Optional[str],
    tree_depth: Optional[int],
    top_p: Optional[float],
) -> None:
    """Setup agent with user inputs and generate README.md file."""
    try:
        conf = ConfigLoader()
        conf.config.api.rate_limit = rate_limit
        llm_api, llm_model = get_environment(api, model)
        conf.config.llm = conf.config.llm.copy(
            update={
                "api": llm_api,
                "base_url": base_url,
                "context_window": context_window,
                "model": llm_model,
                "temperature": temperature,
                "top_p": top_p,
            }
        )
        conf.config.md = conf.config.md.copy(
            update={
                "alignment": alignment,
                "badge_color": badge_color,
                "badge_style": badge_style,
                "emojis": emojis,
                "tree_depth": tree_depth,
                "image": image
                if image
                not in [ImageOptions.CUSTOM.name, ImageOptions.LLM.name]
                else prompt_for_image(image),
            }
        )
        try:
            conf.config.git = GitSettings(repository=repository)
            conf.config.git.host_domain = conf.config.git.repository.host
        except Exception:
            conf.config.git.host_domain = GitHost.LOCAL.name

        asyncio.run(readme_generator(conf, output_file))

    except Exception as exc:
        raise ReadmeGeneratorError(exc, traceback.format_exc()) from exc


async def readme_generator(conf: ConfigLoader, output_file: Path) -> None:
    """Orchestrates the README.md file generation process."""
    repo = conf.config.git.repository
    _logger.info(f"Repository settings validated: {conf.config.git}")
    _logger.info(f"LLM API settings validated: {conf.config.llm}")

    with tempfile.TemporaryDirectory() as temp_dir:
        await clone_repository(repo, temp_dir)
        (
            dependencies,
            raw_files,
        ) = preprocessor(conf, temp_dir)
        _logger.info(f"Dependencies extracted: {dependencies}")
        _logger.info(f"Total files analyzed: {len(raw_files)}")

        async with model_handler(conf).use_api() as llm:
            responses = await llm.batch_request(dependencies, raw_files)
            (
                summaries,
                features,
                overview,
                slogan,
            ) = responses
            conf.config.md.features = conf.config.md.features.format(features)
            conf.config.md.overview = conf.config.md.overview.format(overview)
            conf.config.md.slogan = slogan

        readme_md = MarkdownBuilder(
            conf, dependencies, summaries, temp_dir
        ).build()
        FileHandler().write(output_file, readme_md)

        _logger.info("README generation process completed successfully!")
        _logger.info(f"README.md file saved to: {output_file}")
        _logger.info("Share it @ github.com/eli64s/readme-ai/discussions")
        _logger.info("Thank you for using readme-ai!")



"""Parser for yarn.lock dependency files."""

import re
from typing import List

from readmeai.core.parsers import BaseFileParser


class YarnLockParser(BaseFileParser):
    """Parser for yarn.lock dependency files."""

    def parse(self, content: str) -> List[str]:
        """Extracts package names from a yarn.lock file."""
        try:
            return re.findall(r"(\S+)(?=@)", content)
        except re.error as exc:
            return self.handle_parsing_error(f"yarn.lock: {str(exc)}")



"""Utilities to remove default emojis from markdown content."""

import re
from typing import List

EMOJI_PATTERN = re.compile(
    pattern="["
    "\U0001f600-\U0001f64f"  # emoticons
    "\U0001f300-\U0001f5ff"  # symbols & pictographs
    "\U0001f680-\U0001f6ff"  # transport & map symbols
    "\U0001f700-\U0001f77f"  # alchemical symbols
    "\U0001f780-\U0001f7ff"  # Geometric Shapes Extended
    "\U0001f800-\U0001f8ff"  # Supplemental Arrows-C
    "\U0001f900-\U0001f9ff"  # Supplemental Symbols and Pictographs
    "\U0001fa00-\U0001fa6f"  # Chess Symbols
    "\U0001fa70-\U0001faff"  # Symbols and Pictographs Extended-A
    "\U00002702-\U000027b0"  # Dingbats
    "\U000024c2-\U0001f251"  # flags (iOS)
    "]+",
    flags=re.UNICODE,
)


def remove_emojis(md_content: List[str]) -> List[str]:
    """Removes emojis from the content list."""
    modified_content = []

    for section in md_content:
        lines = section.split("\n")
        for index, line in enumerate(lines):
            if (
                line.startswith("#")
                or "Table of Contents" in section
                or "Quick Links" in section
            ):
                lines[index] = EMOJI_PATTERN.sub("", line)
        modified_content.append("\n".join(lines))

    return modified_content


def split_markdown_headings(markdown_text: str) -> dict:
    """
    Splits a markdown document by level 2 headings into separate sections.
    """
    sections = re.split(r"(?m)^## ", markdown_text)
    split_sections = {}

    for section in sections:
        if section.strip():
            heading = section.split("\n", 1)[0].strip()
            file_name = heading.lower().replace(" ", "_") + ".md"
            content = "## " + section if split_sections else section
            content = content.rstrip("\n")
            split_sections[file_name] = content

    return split_sections


def update_heading_names(md_contents: dict) -> dict:
    """
    Updates dict keys by removing leading emojis, underscores, and spaces.
    """
    updated_md_headings = {}

    for key in md_contents:
        new_key = re.sub(r"^[^\w]+", "", key).lstrip("_")
        if re.match(r"^<.*>.md$", new_key):
            new_key = "header.md"
        updated_md_headings[new_key] = md_contents[key]

    return updated_md_headings



"""Parse package dependencies from go.mod files."""

import re
from typing import List

from readmeai.core.parsers import BaseFileParser


class GoModParser(BaseFileParser):
    """Parser for go.mod files."""

    def __init__(self) -> None:
        """Initializes the handler with given configuration."""
        super().__init__()

    def parse(self, content: str) -> List[str]:
        """Parse the content of a go.mod file."""
        try:
            lines = content.split("\n")
            pattern = r"^\s*([\w\.\-_/]+)\s+v[\w\.\-_/]+"
            regex = re.compile(pattern)
            package_names = (
                regex.match(line.strip()).group(1).split("/")[-1]
                for line in lines
                if regex.match(line.strip())
            )
            return list(package_names)

        except Exception as exc:
            return self.handle_parsing_error(f"go.mod: {str(exc)}")



"""Utility methods for the readme-ai CLI application."""

from __future__ import annotations

import os
from enum import Enum

from readmeai.cli.options import ModelOptions as llms
from readmeai.core.logger import Logger

_logger = Logger(__name__)


class SecretKey(str, Enum):
    """
    Enum class to store the environment variable keys for the LLM API services.
    """

    OLLAMA_HOST = "OLLAMA_HOST"
    OPENAI_API_KEY = "OPENAI_API_KEY"
    VERTEXAI_LOCATION = "VERTEXAI_LOCATION"
    VERTEXAI_PROJECT = "VERTEXAI_PROJECT"


def _set_offline(message: str) -> tuple:
    """Set the LLM service to offline mode."""
    _logger.warning(f"{message}\n\t\t...Generating README without LLM API...")
    return llms.OFFLINE.name, llms.OFFLINE.name


def get_environment(llm_api: str, llm_model: str) -> tuple:
    """Set LLM environment variables based on the specified LLM service."""
    _log_message = "\n\t\t...{} settings NOT FOUND in environment..."

    if llm_api:
        _log_message = "{} settings FOUND in environment!"
        if llm_api == llms.OPENAI.name:
            if SecretKey.OPENAI_API_KEY.value in os.environ:
                _logger.info(_log_message.format("OpenAI"))
                return (
                    llms.OPENAI.name,
                    llm_model if llm_model is not None else "gpt-3.5-turbo",
                )
            else:
                return _set_offline(_log_message.format("OpenAI"))

        elif llm_api == llms.OLLAMA.name:
            if SecretKey.OLLAMA_HOST.value in os.environ:
                _logger.info(_log_message.format("Ollama"))
                return (
                    llms.OLLAMA.name,
                    llm_model if llm_model is not None else "mistral",
                )
            else:
                return _set_offline(_log_message.format("Ollama"))

        elif llm_api == llms.VERTEX.name:
            if (
                SecretKey.VERTEXAI_LOCATION.value in os.environ
                and SecretKey.VERTEXAI_PROJECT.value in os.environ
            ):
                _logger.info(_log_message.format("Goolge Cloud Vertex AI"))
                return (
                    llms.VERTEX.name,
                    llm_model if llm_model is not None else "gemini-pro",
                )

            else:
                return _set_offline(_log_message.format("Vertex AI"))

        elif llm_api == llms.OFFLINE.name:
            return _set_offline("\n\t\t...Offline mode enabled by user...")

    else:
        _log_message = "Running CLI with {} API llm engine..."
        _logger.warning(
            "NO LLM service provided to CLI. Checking environment."
        )

        if SecretKey.OPENAI_API_KEY.value in os.environ:
            _logger.info(_log_message.format("OpenAI"))
            return (
                llms.OPENAI.name,
                llm_model if llm_model is not None else "gpt-3.5-turbo",
            )

        elif SecretKey.OLLAMA_HOST.value in os.environ:
            _logger.info(_log_message.format("Ollama"))
            return (
                llms.OLLAMA.name,
                llm_model if llm_model is not None else "mistral",
            )
        elif (
            SecretKey.VERTEXAI_LOCATION.value in os.environ
            and SecretKey.VERTEXAI_PROJECT.value in os.environ
        ):
            _logger.info(_log_message.format("Vertex AI"))
            return (
                llms.VERTEX.name,
                llm_model if llm_model is not None else "gemini-pro",
            )

        else:
            if llm_api == llms.OFFLINE.name:
                message = "Offline mode enabled by user via CLI."
            else:
                message = (
                    "\n\n\t\t...No LLM API settings found in environment..."
                )
                return _set_offline(message)



"""
Google Cloud's Vertex AI LLM API implementation.
"""

import os
from typing import List, Tuple

import aiohttp
import vertexai
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
from vertexai.preview.generative_models import GenerativeModel

from readmeai.config.settings import ConfigLoader
from readmeai.core.models import BaseModelHandler
from readmeai.models.tokens import token_handler
from readmeai.utils.text_cleaner import clean_response


class VertexAIHandler(BaseModelHandler):
    """Google Cloud Vertex AI LLM API implementation."""

    def __init__(self, config_loader: ConfigLoader) -> None:
        """Initialize GCP Vertex AI API LLM handler."""
        super().__init__(config_loader)
        self._model_settings()

    def _model_settings(self):
        """Initializes the Vertex AI LLM settings."""
        self.location = os.environ.get("VERTEXAI_LOCATION")
        self.project_id = os.environ.get("VERTEXAI_PROJECT")
        self.temperature = self.config.llm.temperature
        self.tokens = self.config.llm.tokens
        self.top_p = self.config.llm.top_p
        vertexai.init(location=self.location, project=self.project_id)
        self.model = GenerativeModel(self.config.llm.model)

    async def _build_payload(self, prompt: str, tokens: int) -> dict:
        """Build payload for POST request to Vertex AI API."""
        return {
            "max_output_tokens": tokens,
            "temperature": self.temperature,
            "top_p": self.top_p,
        }

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=4, max=10),
        retry=retry_if_exception_type(
            (
                aiohttp.ClientError,
                aiohttp.ClientResponseError,
                aiohttp.ClientConnectorError,
            )
        ),
    )
    async def _make_request(
        self,
        index: str,
        prompt: str,
        tokens: int,
        raw_files: List[Tuple[str, str]] = None,
    ) -> Tuple[str, str]:
        """Processes Vertex AI LLM API responses and returns generated text."""
        try:
            prompt = await token_handler(self.config, index, prompt, tokens)

            data = await self._build_payload(prompt, tokens)

            async with self.rate_limit_semaphore:
                response = await self.model.generate_content_async(
                    prompt,
                    generation_config=data,
                )
                content = response.candidates[0].content.parts[0].text
                self._logger.info(f"Response for '{index}':\n{content}")
                return index, clean_response(index, content)

        except (
            aiohttp.ClientError,
            aiohttp.ClientResponseError,
            aiohttp.ClientConnectorError,
        ) as exc:
            self._logger.error(
                f"Error making request to Vertex AI for `{index}`: {exc}"
            )
            return index, self.config.md.placeholder



"""Methods to validate command-line arguments and settings."""

import re
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urlparse, urlsplit

from readmeai._exceptions import GitValidationError
from readmeai.core.logger import Logger
from readmeai.services.git import GitHost

_logger = Logger(__name__)


class GitValidator:
    """Validator class for GitSettings."""

    @classmethod
    def validate_repository(cls, value: Union[str, Path]) -> Union[str, Path]:
        """Validate the repository URL or path."""
        value_str = str(value)
        if (
            any(service.value in value_str for service in GitHost) is False
            and not Path(value_str).is_dir()
        ):
            raise GitValidationError(value)

        if isinstance(value, str):
            path = Path(value)
            if path.is_dir():
                return value
            try:
                parsed_url = urlparse(value)
                if parsed_url.scheme in ["http", "https"] and any(
                    service in parsed_url.netloc for service in GitHost
                ):
                    return value

            except Exception as exc:
                _logger.error(f"Error validating repository: {exc}")
                raise GitValidationError(value) from exc

        elif isinstance(value, Path) and value.is_dir():
            return value

        raise GitValidationError(value)

    @classmethod
    def validate_full_name(cls, value: Optional[str], values: dict) -> str:
        """Validator for getting the full name of the repository."""
        url_or_path = values.get("repository")

        path = (
            url_or_path if isinstance(url_or_path, Path) else Path(url_or_path)
        )
        if path.exists():
            return str(path.name)

        patterns = {
            GitHost.GITHUB: r"https?://github.com/([^/]+)/([^/]+)",
            GitHost.GITLAB: r"https?://gitlab.com/([^/]+)/([^/]+)",
            GitHost.BITBUCKET: r"https?://bitbucket.org/([^/]+)/([^/]+)",
        }

        for _, pattern in patterns.items():
            match = re.match(pattern, url_or_path)
            if match:
                user_name, repo_name = match.groups()
                return f"{user_name}/{repo_name}"

        raise GitValidationError(url_or_path)

    @classmethod
    def set_host(cls, value: Optional[str], values: dict) -> str:
        """Sets the Git service host from the repository provided."""
        repo = values.get("repository")
        if isinstance(repo, Path) or (
            isinstance(repo, str) and Path(repo).is_dir()
        ):
            return GitHost.LOCAL

        parsed_url = urlparse(str(repo))
        for service in GitHost:
            if service in parsed_url.netloc:
                return service.split(".")[0]

        return GitHost.LOCAL

    @classmethod
    def set_name(cls, value: Optional[str], values: dict) -> str:
        """Sets the repository name from the repository provided."""
        repo = values.get("repository")
        if isinstance(repo, Path):
            return repo.name
        elif isinstance(repo, str):
            parsed_url = urlsplit(repo)
            name = parsed_url.path.split("/")[-1]
            return name.removesuffix(".git")
        return "n/a"



"""Abstract base class for dependency file parsers."""

from abc import ABC, abstractmethod
from typing import List

from readmeai.core.logger import Logger


class BaseFileParser(ABC):
    """Abstract base class for dependency file parsers."""

    def __init__(self) -> None:
        """Initializes the handler with given configuration."""
        self._logger = Logger(__name__)

    @abstractmethod
    def parse(self, content: str) -> List[str]:
        """Parses content of dependency file and returns list of dependencies."""
        ...

    def log_error(self, message: str):
        """Logs error message when parsing fails."""
        self._logger.error(f"Error parsing dependency file {message}")

    def handle_parsing_error(self, error: Exception) -> List[str]:
        """Standardized error handling for parsing exceptions."""
        self.log_error(str(error))
        return []
